unit GEPanel;

interface

uses
  messages,Windows,SysUtils, Classes,Controls,StdCtrls,ExtCtrls,
  ShellApi,Dialogs,ActiveX,KEYHOLELib_TLB,Registry,Graphics,jpeg;

type
  TGEPanel = class(TCustomPanel)
  private
    { Private declarations }
  protected
    { Protected declarations }
    geGrabbed:boolean;
    hGE:HWND;
    hRender:HWND;
    hParentRender:HWND;
    hPrevRender:HWND;
    hNavigation:HWND;
    FConnected:boolean;
    timer:TTimer;
    FNavigation:boolean;
    FKHInterface: TKHInterface;
    bm : TBitMap;
    function GetKHInterface:IKHInterface;
    procedure SetNavigation(v:boolean);
    procedure SetGoogleEarth;
    procedure checkGE(Sender:TObject);
    procedure Readregistry;
    procedure setlayer(name:String;v:integer);
  public
    xlat,ylon:Double;
    layers:TStringList;
    procedure Paint; override;
    procedure Loaded; override;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMDestroy(var Message: TWMDestroy); message WM_DESTROY;
    procedure WMMouseActivate(var Message: TWMMouseActivate);message WM_MOUSEACTIVATE;
    procedure WMMousePos(var Message: TMessage); message WM_USER+1000;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    function  GetAzimuth:Double;
    procedure SetAzimuth(v:Double);
    function  GetRange:Double;
    procedure SetRange(v:Double);
    function  GetTilt:Double;
    procedure SetTilt(v:Double);
    procedure ScreenShot;
    procedure SaveKMLToFile(fn:String);
    procedure RestoreGEWindow;
    constructor Create(Owner: TComponent); override;
    destructor Destroy;Override;
    { Public declarations }
    property Connected:boolean read FConnected;
    property KHInterface: IKHInterface read GetKHInterface;
    property Navigation:boolean read FNavigation write SetNavigation;
    property Azimuth:Double read GetAzimuth write SetAzimuth;
    property Range:Double read Getrange write SetRange;
    property Tilt:Double read GetTilt write SetTilt;
  published
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property BevelOuter;
    property BorderWidth;
    property BorderStyle;
    property Caption;
    property Color;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor;
    property ParentFont;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDrag;
    { Published declarations }
  end;

procedure Register;

implementation

var
hGlobalRender:HWND;

constructor TGEPanel.Create(Owner: TComponent);
begin
   inherited Create(Owner);
   DoubleBuffered := false;
   geGrabbed:=false;
   hrender:=0;
   FConnected:=false;
   FNavigation:=false;
   OleInitialize(nil);
   Font.Size:=12;
   Font.Color:=$ffffff;
   Color:=$000000;
   Caption:='';
   Align:=alClient;
   BevelOuter:=bvNone;
   timer:=TTimer.Create(self);
   timer.OnTimer:=checkGE;
   timer.Interval:=200;
   timer.Enabled:=true;
   parentBackground:=true;
   parentBackground:=false;
   layers:=TStringList.Create;
   //setlayer('Italy Roads',1);
   //Readregistry;
end;

destructor TGEPanel.Destroy;
Begin
   layers.Free;
   if FConnected then begin
      if not (csDesigning in ComponentState) then begin
         try
            //RestoreGEWindow;
            FConnected:=FKHInterface.DefaultInterface<>nil;
            if Connected then begin
               //FKHInterface.Disconnect;
               //FKHInterface.QuitApplication;
            end;
         except
         end;
      end;
   end;
   OleUninitialize;
   Inherited;
end;

procedure TGEPanel.RestoreGEWindow;
var
   r:TRect;
begin
   //SWP_NOSIZE	Retains the current size (ignores the cx and cy parameters).
   Windows.SetParent(hrender,hparentRender);
   Windows.getClientrect(hparentRender,r);
   Windows.SetWindowPos(hrender,hPrevRender,r.left,r.top,r.Right-r.left,r.Bottom-r.top,0);//SWP_NOREPOSITION	or SWP_NOSIZE
   Windows.MoveWindow(hrender,r.left,r.top,r.Right-r.left,r.Bottom-r.top,true);
   SetWindowText(hRender,'Render');
   ShowWindow(hge,SW_SHOW);
end;

procedure TGEPanel.CMEnter(var Message: TCMEnter);
begin
   inherited;
   if not (csDesigning in ComponentState) then begin
      if Connected and geGrabbed then begin
         if hrender<>0 then begin
            Windows.SetFocus(hrender);
         end;
      end;
   end;
end;

procedure TGEPanel.checkGE(Sender:TObject);
var
   hGE:HWND;
begin
   if not (csDesigning in ComponentState) then begin
      hGE:=FindWindow('QWidget','Google Earth');
      if hGE<>0 then begin
         ShowWindow(hge,SW_HIDE);
         if Connected and geGrabbed then begin
            timer.Enabled:=false;
         end;
      end;
   end;
end;

procedure TGEPanel.SetNavigation(v:boolean);
begin
   if not (csDesigning in ComponentState) then begin
      if Connected and geGrabbed then begin
         if hrender<>0 then begin
            Windows.SetFocus(hrender);
         end;
      end;
   end;
end;

procedure TGEPanel.WMDestroy(var Message: TWMDestroy);
begin
   RestoreGEWindow;
   inherited;
end;

procedure TGEPanel.WMMouseActivate(var Message: TWMMouseActivate);
var
   pt:TPoint;
begin
   inherited;
   if not (csDesigning in ComponentState) then begin
      if Connected and geGrabbed then begin
         if hrender<>0 then begin
            Windows.SetFocus(hrender);
         end;
         pt:=Mouse.CursorPos;
         pt:=ScreenToClient(pt);
         //PostMessage(handle,WM_user+1000,makelong(pt.X,pt.Y),0);
      end;
   end;
end;

procedure TGEPanel.WMMousePos(var Message: TMessage);
var
   sa:PSafeArray;
   index,x,y:integer;
   x1,y1:Double;
begin
   x:=LoWord(message.wparam);
   y:=HiWord(message.wparam);
   x1:=(x-Width/2)/Width;
   y1:=(y-height/2)/height;
   sa:=FKHInterface.getPointOnTerrainFromScreenCoords(x1,y1);
   index:=0;
   SafeArrayGetElement(sa,index,xlat);
   //x:=v;
   index:=1;
   SafeArrayGetElement(sa,index,ylon);
   PostMessage(parent.handle,WM_user+1000,0,0);
   //y:=v;
end;

function TGEPanel.GetKHInterface;
begin
   result:=FKHInterface.DefaultInterface;
end;

procedure TGEPanel.Loaded;
var
   s:String;
   i:integer;
begin
   inherited;
   if not (csDesigning in ComponentState) then begin
      try
         FKHInterface:=TKHInterface.Create(self);
         FKHInterface.Connect;
      except
         ShellExecute(Handle,'open', 'http://dl.google.com/earth/GoogleEarth.exe',nil,nil, SW_SHOWNORMAL) ;
         i:=0;
         while i<300 do begin
            Sleep(1000);
            try
               FKHInterface.Connect;
               break;
            except
            end;
            i:=i+1;
         end;
      end;
      FConnected:=FKHInterface.DefaultInterface<>nil;
      if FConnected then begin
         //FKHInterface.SetRenderWindowSize(100,100);
         //OutputDebugString('GE connected');
      end else begin
         //OutputDebugString('GE not yet connected');
      end;
      (*if FConnected then begin
         while hrender=0 do begin
            SetGoogleEarth;
            Application.ProcessMessages;
         end;
      end;*)

   end;
end;

procedure TGEPanel.WMSize(var Message: TWMSize);
begin
   inherited;
   if not (csDesigning in ComponentState) then begin
      if FConnected then begin
         if not geGrabbed then begin
            SetGoogleEarth;
         end;
         if geGrabbed then begin
            if hrender<>0 then begin
               MoveWindow(hrender,0,0,Width,Height,true);
            end;
         end;
      end;
   end;
end;

procedure TGEPanel.Paint;
var
   HP,hn,hc,h:HWND;
   rr,ll:integer;
begin
   inherited;
   if not (csDesigning in ComponentState) then begin
      if FConnected then begin
         if geGrabbed then begin
            hp:=Windows.GetParent(hrender);
            if hp<>handle then begin
               SetGoogleEarth;
            end;
            if hrender<>0 then begin
               MoveWindow(hrender,0,0,Width,Height,true);
            end;
         end;
         if not geGrabbed then begin
            SetGoogleEarth;
         end;
      end;
   end;
end;

function  TGEPanel.GetAzimuth:Double;
var
   vi:IKHViewInfo;
begin
   result:=0;
   if FConnected then begin
      vi:=FKHInterface.currentView(0);
      result:=vi.azimuth;
   end;
end;

procedure TGEPanel.SetAzimuth(v:Double);
var
   vi:IKHViewInfo;
begin
   if FConnected and (FKHInterface<>nil) and geGrabbed then begin
      try
         vi:=FKHInterface.currentView(0);
         if vi<>nil then begin
            vi.azimuth:=v;
            FKHInterface.setView(vi,0,1);
         end;
      except
      end;
   end;
end;

function  TGEPanel.GetRange:Double;
var
   vi:IKHViewInfo;
begin
   result:=0;
   if FConnected then begin
     vi:=FKHInterface.currentView(0);
      result:=vi.Range;
   end;
end;

procedure TGEPanel.SetRange(v:Double);
var
   vi:IKHViewInfo;
begin
   if FConnected and (FKHInterface<>nil) and geGrabbed then begin
      try
         vi:=FKHInterface.currentView(0);
         if vi<>nil then begin
            vi.Range:=v;
            FKHInterface.setView(vi,0,1);
         end;
      except
      end;
   end;
end;

function  TGEPanel.GetTilt:Double;
var
   vi:IKHViewInfo;
begin
   result:=0;
   if FConnected then begin
      vi:=FKHInterface.currentView(0);
      result:=vi.Tilt;
   end;
end;

procedure TGEPanel.SetTilt(v:Double);
var
   vi:IKHViewInfo;
begin
   if FConnected and (FKHInterface<>nil) and geGrabbed then begin
      try
         vi:=FKHInterface.currentView(0);
         if vi<>nil then begin
            vi.Tilt:=v;
            FKHInterface.setView(vi,0,1);
         end;
      except
      end;
   end;
end;


procedure TGEPanel.SetGoogleEarth;
var
   hn,hc:HWND;
   ptr:pchar;
begin
   //use spyxx.exe from mssdk to see windows tree structure
   hGE:=FindWindow('QWidget','Google Earth');
   getmem(ptr,128);
   if hGlobalRender<>0 then begin
      hrender:=hGlobalRender;
   end else if hGE<>0 then begin
      ShowWindow(hge,SW_SHOWMINIMIZED);
      MoveWindow(hge,0,0,100,100,true);
      hc:=GetWindow(hGE,GW_CHILD);
      hc:=GetWindow(hc,GW_HWNDNEXT);
      hn:=GetWindow(hc,GW_HWNDNEXT);//qtcentralwindow
         hn:=GetWindow(hn,GW_CHILD);//MAINWINDOWHSPLITTER
            hn:=GetWindow(hn,GW_CHILD);//qtsplithandle
            hn:=GetWindow(hn,GW_HWNDNEXT);//CENTERPANELVSPLITTER
               hn:=GetWindow(hn,GW_CHILD);//WEBWINDOWFRAME
               hNavigation:=GetWindow(hn,GW_HWNDNEXT);//CENTERPANELVSPLITTER
               hParentRender:=GetWindow(hNavigation,GW_HWNDNEXT);//CENTERPANELVSPLITTER
                  hPrevRender:=GetWindow(hParentRender,GW_CHILD);//WEBWINDOWFRAME
                  getWindowText(hPrevRender,ptr,127);
                  hn:=GetWindow(hPrevRender,GW_HWNDNEXT);//CENTERPANELVSPLITTER
                  getWindowText(hn,ptr,127);
                  hn:=GetWindow(hn,GW_CHILD);
                  getWindowText(hn,ptr,127);
                  hn:=GetWindow(hn,GW_CHILD);
                  getWindowText(hn,ptr,127);
                  hrender:=GetWindow(hn,GW_CHILD);
                  getWindowText(hrender,ptr,127);
                  hParentRender:=GetParent(hRender);
                  hGlobalRender:=hRender;

   end;
   Windows.SetParent(hrender,handle);
   geGrabbed:=true;
   MoveWindow(hrender,0,0,Width,Height,true);
   ShowWindow(hge,SW_HIDE);
   freemem(ptr,128);
end;

procedure TGEPanel.Readregistry;
const
   Key  : String = '\Software\Google\CommonSettings\Layers\';
var
   Data : Array of Byte;
   L    : TStringList;
   S    : String;
   sz   : Word;
   I,J  : Word;
   dw:dword;
   r:TRegistry;
begin
   R := TRegistry.Create;
   L := TStringList.Create;
   with R do begin
      RootKey := HKEY_CURRENT_USER;
      OpenKey(Key, False);
      GetValueNames(L);
      If L.Count > 0 Then begin
         for I := 0 to L.Count - 1 do        begin
            dw:=0;
            case GetDataType(L[I]) of
               rdString,
               rdExpandString :
                  S := '"' + ReadString(L[I]) + '"';
               rdInteger : begin
                  dw:=ReadInteger(L[I]);
                  S := IntToStr(dw);
               end;
               rdBinary :          begin
                  sz := GetDataSize(L[I]);
                  SetLength(Data, sz);
                  ReadBinaryData(L[I], Data[0], sz);
                  S := '';
                  for J := 0 to sz - 1 do            begin
                     S := S + Format('%2x',[Data[I]]);
                  end;
               end;
               rdUnknown :
                 S := 'Unknown';
            end;
            layers.AddObject(l[i],TObject(dw));
         end;
      end;
      Free;
   end;
   L.Free;
end;

procedure TGEPanel.setLayer(name:String;v:integer);
const
   Key  : String = '\Software\Google\CommonSettings\Layers\';
var
   //Data : Array of Byte;
   L    : TStringList;
   S    : String;
   sz   : Word;
   I,J  : Word;
   dw:dword;
   r:TRegistry;
begin
   R := TRegistry.Create;
   L := TStringList.Create;
   with R do begin
      RootKey := HKEY_CURRENT_USER;
      OpenKey(Key, False);
      GetValueNames(L);
      If L.Count > 0 Then begin
         for I := 0 to L.Count - 1 do        begin
            if Pos(name,l[i])>0 then begin
               WriteInteger(l[i],v);
            end;
         end;
      end;
      Free;
   end;
   L.Free;
end;

procedure TGEPanel.ScreenShot;
var
   dc: HDC;
   lpPal : PLOGPALETTE;
   jp:TJPEGImage;
   sd:TSaveDialog;
begin
   bm := TBitMap.Create;
   bm.Width := Width;
   bm.Height := Height;
   {get the screen dc}
   dc := GetDc(Handle);
   if (dc = 0) then exit;

   {do we have a palette device?}
   if (GetDeviceCaps(dc, RASTERCAPS) AND RC_PALETTE = RC_PALETTE) then begin
      {allocate memory for a logical palette}
      GetMem(lpPal, sizeof(TLOGPALETTE) + (255 * sizeof(TPALETTEENTRY)));
      {zero it out to be neat}
      FillChar(lpPal^, sizeof(TLOGPALETTE) + (255 * sizeof(TPALETTEENTRY)), #0);
      {fill in the palette version}
      lpPal^.palVersion := $300;
      {grab the system palette entries}
      lpPal^.palNumEntries := GetSystemPaletteEntries(dc, 0, 256, lpPal^.palPalEntry);
      if (lpPal^.PalNumEntries <> 0) then begin
          {create the palette}
          bm.Palette := CreatePalette(lpPal^);
      end;
      FreeMem(lpPal, sizeof(TLOGPALETTE) + (255 * sizeof(TPALETTEENTRY)));
   end;

   {copy from the screen to the bitmap}
   BitBlt(bm.Canvas.Handle, 0, 0, Width, Height, Dc, 0, 0, SRCCOPY);
   {release the screen dc}
   ReleaseDc(Handle, dc);

   sd:=TSaveDialog.Create(self);
   sd.DefaultExt := 'jpg';
   sd.Filter := 'Jpeg Images (*.jpg)';
   if sd.execute then begin
      jp:=TJPEGImage.Create;
      jp.assign(bm);
      jp.SaveToFile(ChangeFileext(sd.Filename,'.jpg'));
      jp.Free;
      //SaveKMLToFile(ChangeFileext(sd.Filename,'.kml'));
   end;
   sd.Free;
   bm.Free;
end;

procedure TGEPanel.SaveKMLToFile(fn:String);
var
   ssDoc,Srange,STilt,Slat,Slon,s:String;
   vi:IKHViewInfo;
   sl:TStringList;
begin
   if FConnected then begin
      vi:=FKHInterface.currentView(0);

      SLat:=FormatFloat('##.####',vi.latitude);
      SLon:=FormatFloat('###.####',vi.longitude);
      Slat:=StringReplace(slat,',','.',[rfReplaceAll]);
      Slon:=StringReplace(slon,',','.',[rfReplaceAll]);

      SRange:=FormatFloat('0.####',range);
      STilt:=FormatFloat('0.####',Tilt);
      SRange:=StringReplace(SRange,',','.',[rfReplaceAll]);
      STilt:=StringReplace(STilt,',','.',[rfReplaceAll]);
      //SHeading:=FormatFloat('0.####',Azimouth);

      sl:=TStringList.Create;

      sl.Add('<?xml version="1.0" encoding="UTF-8"?>');
      sl.Add('<kml xmlns="http://earth.google.com/kml/2.0">');
      sl.Add('<Placemark>');
      sl.Add('<LookAt>');
      sl.Add('<longitude>'+Slon+'</longitude>');
      sl.Add('<latitude>'+Slat+'</latitude>');
      sl.Add('<range>'+Srange+'</range>');
      sl.Add('<tilt>'+STilt+'</tilt>');
      //sl.Add('<heading>'+SAzimuth+'</heading>');
      sl.Add('</LookAt>');

      sl.Add('<styleUrl>root://styleMaps#default+nicon=0x307+hicon=0x317</styleUrl>');
      sl.Add('<Style>');
      sl.Add('<IconStyle>');
      sl.Add('<Icon>');
      sl.Add('<href>root://icons/palette-3.png</href>');
      sl.Add('<x>96</x>');
      sl.Add('<y>128</y>');
      sl.Add('<w>32</w>');
      sl.Add('<h>32</h>');
      sl.Add('</Icon>');
      sl.Add('</IconStyle>');
      sl.Add('</Style>');

      sl.Add('<Point>');
      sl.Add('<coordinates>');
      sl.Add(sLat+',');
      sl.Add(sLon+',');
      sl.Add('0');
      sl.Add('</coordinates>');
      sl.Add('</Point>');

      sl.Add('</Placemark>');
      sl.Add('</kml>');
      sl.SaveToFile(fn);
      sl.Free;
   end;
end;

procedure Register;
begin
  RegisterComponents('Internet', [TGEPanel]);
end;

end.
